/*
 * The contents of this file are subject to the AOLserver Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://aolserver.lcs.mit.edu/.
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is AOLserver Code and related documentation
 * distributed by AOL.
 * 
 * The Initial Developer of the Original Code is America Online,
 * Inc. Portions created by AOL are Copyright (C) 1999 America Online,
 * Inc. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms
 * of the GNU General Public License (the "GPL"), in which case the
 * provisions of GPL are applicable instead of those above.  If you wish
 * to allow use of your version of this file only under the terms of the
 * GPL and not to allow others to use your version of this file under the
 * License, indicate your decision by deleting the provisions above and
 * replace them with the notice and other provisions required by the GPL.
 * If you do not delete the provisions above, a recipient may use your
 * version of this file under either the License or the GPL.
 */


/*
 * dirindex.c --
 *
 *      Get page possibly from a file cache, or call a directory
 *      index function.  This file is derived from fastpath.c.
 *
 */

static const char *RCSID = "@(#) $Header: /u/cvsroot/dqd/aolserver3/contrib/dqd_dirindex/dirindex.c,v 1.4 2000/01/28 07:03:48 mayoff Exp $, compiled: " __DATE__ " " __TIME__;

#include "ns.h"
#include <sys/stat.h>
#ifndef WIN32
#include <sys/mman.h>
#endif

int Ns_ModuleVersion = 1;

/*
 * The following constants are defined for this file
 */

#ifndef MAP_FAILED
#define MAP_FAILED ((void *) (-1))
#endif

/*
 * The following structure defines the key to identify a 
 * in the file cache.
 */

typedef struct {
    dev_t dev;
    ino_t ino;
} Key;

/*
 * The following structure defines the contents of a file
 * stored in the file cache.
 */

typedef struct {
    time_t mtime;
    int size;
    int refcnt;
    char bytes[1];	/* really char bytes[size]; file contents is copied
			 * here */
} File;

/*
 * Local functions defined in this file
 */

static int FastGet(void *ignored, Ns_Conn *conn);
static Ns_Callback FreeEntry;
static void DecrEntry(File *);
static int UrlIs(char *server, char *url, int dir);
static void GetConfig(const char *server, const char *module);
static Ns_OpProc TclDirIndexer;

/*
 * Static variables defined in this file
 */

static Ns_Cache 	*cachePtr;	/* the file cache */
static Ns_ModLogHandle   dirIndexLog;
static Ns_OpProc        *dirHandlerPtr;
static void             *dirHandlerContext;

static struct {
    char **dirv;
    int dirc;
    int mmap;
    int cache;
    int cachesize;
    int cachemaxentry;
    char *tclCommand;
} config;

/*
 *==========================================================================
 * API functions
 *==========================================================================
 */


/*
 *----------------------------------------------------------------------
 * Ns_SetDirHandlerProc --
 *
 *      Set pointer to custom routine that will handle requests that
 *      map to filesystem directories.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

void
Dqd_SetDirHandlerProc(Ns_OpProc *procPtr, void *context)
{
    dirHandlerPtr = procPtr;
    dirHandlerContext = context;
}

/* 
 *==========================================================================
 * Exported functions
 *==========================================================================
 */


/*
 *----------------------------------------------------------------------
 * NsInitFastpath --
 *
 *      Initialization function called at server start up.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      File cache may be created.
 *
 *----------------------------------------------------------------------
 */

int
Ns_ModuleInit(char *server, char *module)
{
    Ns_ModLogRegister(module, &dirIndexLog);    

    GetConfig(server, module);

    if (config.cache) {
	cachePtr = Ns_CacheCreateSz(module, sizeof(Key) / sizeof(int),
				    config.cachesize, FreeEntry);
    }

    if (config.tclCommand != NULL) {
	Dqd_SetDirHandlerProc(TclDirIndexer, NULL);
    }

    Ns_RegisterRequest(NULL, "GET", "/", FastGet, NULL, NULL, 0);
    Ns_RegisterRequest(NULL, "HEAD", "/", FastGet, NULL, NULL, 0);
    Ns_RegisterRequest(NULL, "POST", "/", FastGet, NULL, NULL, 0);

    return NS_OK;
}


/*
 *==========================================================================
 * Static functions
 *==========================================================================
 */


/*
 *----------------------------------------------------------------------
 * GetConfig --
 *
 *	Initialize config from the configuration file.
 *
 * Results:
 *      config struct initialized.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static void
GetConfig(const char *server, const char *module)
{
    Ns_DString ds;
    char *section;
    char *serverSection;
    char *dirlist;

    Ns_DStringInit(&ds);

    serverSection = Ns_StrDup(Ns_DStringVarAppend(&ds,
	"ns/server/", server, NULL));
    section = Ns_DStringVarAppend(&ds, "/fastpath", NULL);

    if (!Ns_ConfigGetBool(section, "mmap", &config.mmap))
	config.mmap = NS_FALSE;

    if (!Ns_ConfigGetBool(section, "cache", &config.mmap))
	config.cache = NS_TRUE;

    if (config.cache) {
	if (!Ns_ConfigGetInt(section, "cachemaxsize", &config.cachesize))
	    config.cachesize = 5000 * 1024;
	if (!Ns_ConfigGetInt(section, "cachemaxentry", &config.cachemaxentry))
	    config.cachemaxentry = 8192;
    }

    dirlist = Ns_ConfigGetValue(section, "directoryfile");
    if (dirlist == NULL)
	dirlist = Ns_ConfigGetValue(serverSection, "directoryfile");
    if (dirlist != NULL) {
	int i;
	char *p;
	dirlist = Ns_StrDup(dirlist);
	for (config.dirc = 1, p = dirlist; *p; p++) {
	    if (*p == ',')
		config.dirc++;
	}
	config.dirv = Ns_Calloc(config.dirc,
	    sizeof *config.dirv);

	p = dirlist;
	i = 0;
	while (1) {
	    config.dirv[i++] = p;
	    p = strchr(p, ',');
	    if (p == NULL)
		break;
	    *p++ = '\000';
	}
    }

    config.tclCommand = Ns_ConfigGetValue(section, "tclCommand");
    if (config.tclCommand == NULL)
	config.tclCommand = "dirindex";

    Ns_DStringFree(&ds);
    Ns_Free(serverSection);
}


/*
 *----------------------------------------------------------------------
 * FastGetRestart --
 *
 *	Construct the full URL and redirect internally.
 *
 * Results:
 *      See Ns_ConnRedirect().
 *
 * Side effects:
 *      See Ns_ConnRedirect().
 *
 *----------------------------------------------------------------------
 */

static int
FastGetRestart(Ns_Conn *conn, char *page)
{
    int		    status;
    Ns_DString      ds;

    Ns_DStringInit(&ds);
    Ns_MakePath(&ds, conn->request->url, page, NULL);
    status = Ns_ConnRedirect(conn, Ns_DStringValue(&ds));
    Ns_DStringFree(&ds);

    return status;
}


/*
 *----------------------------------------------------------------------
 * FastGet --
 *
 *      Return the contents of a URL.
 *
 * Results:
 *      Return NS_OK for success or NS_ERROR for failure.
 *
 * Side effects:
 *      Contents of file may be cached in file cache.
 *
 *----------------------------------------------------------------------
 */

static int
FastGet(void *ignored, Ns_Conn *conn)
{
    Ns_DString      ds;
    char           *type, *p;
    char	   *url = conn->request->url;
    int             retval, fd, new, nread, i;
    struct stat	    st;
    File	   *filePtr;
    Key		    key;
    Ns_Entry	   *entPtr;

    Ns_DStringInit(&ds);
    if (Ns_UrlToFile(&ds, nsServer, url) != NS_OK) {
	goto notfound;
    }

    if (stat(Ns_DStringValue(&ds), &st) != 0) {
	if (errno != ENOENT) {
	    Ns_ModLog(Error, dirIndexLog, "stat(%s) failed: %s",
		      Ns_DStringValue(&ds), strerror(errno));
	}
	goto notfound;
    }

    if (S_ISREG(st.st_mode)) {
	Ns_ConnSetLastModifiedHeader(conn, &st.st_mtime);

    	if (Ns_ConnModifiedSince(conn, st.st_mtime) == NS_FALSE) {
	    retval = Ns_ReturnNotModified(conn);
	    goto done;
	}
    	type = Ns_GetMimeType(Ns_DStringValue(&ds));

	if (config.cache == NS_FALSE || st.st_size > config.cachemaxentry) {

	    /*
	     * Caching is disabled or the entry is too large for the cache
	     * so just open, mmap, and send the content directly.
	     */

    	    fd = open(Ns_DStringValue(&ds), O_RDONLY);
    	    if (fd < 0) {
	        Ns_ModLog(Warning, dirIndexLog, "open(%s) failed: %s",
			  Ns_DStringValue(&ds), strerror(errno));
	     	goto notfound;
	    }

#ifdef WIN32
            retval = Ns_ConnReturnOpenFd(conn, 200, type, fd, st.st_size);
	    close(fd);
#else
	    if (config.mmap && (p = mmap(0, st.st_size, PROT_READ, MAP_SHARED,
			  fd, 0)) != MAP_FAILED) {
		close(fd);
            	retval = Ns_ConnReturnData(conn, 200, p, st.st_size, type);
		munmap(p, st.st_size);
	    } else {
            	retval = Ns_ConnReturnOpenFd(conn, 200, type, fd, st.st_size);
	    	close(fd);
	    }
#endif

	} else {

	    /*
	     * Search for an existing cache entry for this file, validating 
	     * the contents against the current file mtime and size.
	     */

	    filePtr = NULL;
	    key.dev = st.st_dev;
	    key.ino = st.st_ino;
	    Ns_CacheLock(cachePtr);
	    entPtr = Ns_CacheCreateEntry(cachePtr, (char *) &key, &new);
	    if (!new) {
		while (entPtr != NULL &&
		       (filePtr = Ns_CacheGetValue(entPtr)) == NULL) {
		    Ns_CacheWait(cachePtr);
		    entPtr = Ns_CacheFindEntry(cachePtr, (char *) &key);
		}
		if (filePtr != NULL &&
			(filePtr->mtime != st.st_mtime ||
			 filePtr->size != st.st_size)) {
		    Ns_CacheUnsetValue(entPtr);
		    new = 1;
		}
	    }
	    if (new) {

		/*
		 * Read and cache new or invalidated entries in one big chunk.
		 */
		
		Ns_CacheUnlock(cachePtr);
		fd = open(Ns_DStringValue(&ds), O_RDONLY);
		if (fd < 0) {
	    	    filePtr = NULL;
	            Ns_ModLog(Warning, dirIndexLog,
			      "open(%s) failed: %s", Ns_DStringValue(&ds),
			      strerror(errno));
		} else {
		    filePtr = ns_malloc(sizeof(File) + st.st_size);
		    filePtr->refcnt = 1;
		    filePtr->size = st.st_size;
		    filePtr->mtime = st.st_mtime;
		    nread = read(fd, filePtr->bytes, filePtr->size);
		    close(fd);
		    if (nread != filePtr->size) {
	    	    	Ns_ModLog(Warning, dirIndexLog,
			      "read(%s) failed: %s", Ns_DStringValue(&ds),
			      strerror(errno));
			ns_free(filePtr);
			filePtr = NULL;
		    }
		}
		Ns_CacheLock(cachePtr);
		if (filePtr != NULL) {
		    Ns_CacheSetValueSz(entPtr, filePtr, filePtr->size);
		} else {
		    Ns_CacheDeleteEntry(entPtr);
		}
		Ns_CacheBroadcast(cachePtr);
	    }
	    if (filePtr != NULL) {
		++filePtr->refcnt;
		Ns_CacheUnlock(cachePtr);
            	retval = Ns_ConnReturnData(conn, 200, filePtr->bytes,
				filePtr->size, type);
		Ns_CacheLock(cachePtr);
		DecrEntry(filePtr);
	    }
	    Ns_CacheUnlock(cachePtr);
	    if (filePtr == NULL) {
		goto notfound;
	    }
	}

    } else if (S_ISDIR(st.st_mode)) {
	struct stat dir_st = st;

	/*
	 * A directory URL must end with a slash to make
	 * relative links work.
	 */
	if (url[strlen(url) - 1] != '/') {
	    Ns_DStringTrunc(&ds, 0);
	    Ns_DStringVarAppend(&ds, url, "/", NULL);
	    retval = Ns_ReturnRedirect(conn, Ns_DStringValue(&ds));
	    goto done;
	}

	/*
	 * For directories, search for a matching directory file and restart
	 * the connection if found.
	 */
	if (config.dirc > 0) {
	    int dirLength;

	    Ns_DStringTrunc(&ds, 0);
	    if (Ns_UrlToFile(&ds, nsServer, url) != NS_OK) {
		goto notfound;
	    }
	    dirLength = Ns_DStringLength(&ds);

	    for (i = 0; i < config.dirc; ++i) {
		Ns_DStringTrunc(&ds, dirLength);
		Ns_DStringVarAppend(&ds, "/", config.dirv[i], NULL);
		if ((stat(Ns_DStringValue(&ds), &st) == 0)
		    && S_ISREG(st.st_mode))
		{
		    retval = FastGetRestart(conn, config.dirv[i]);
		    goto done;
		}
	    }
	}

	if (dirHandlerPtr != NULL) {
	    Ns_ConnSetLastModifiedHeader(conn, &dir_st.st_mtime);

	    if (Ns_ConnModifiedSince(conn, dir_st.st_mtime) == NS_FALSE) {
		retval = Ns_ReturnNotModified(conn);
		goto done;
	    }

	    retval = (*dirHandlerPtr)(dirHandlerContext, conn);
	    goto done;
	}

        retval = Ns_ReturnForbidden(conn);
    } else {
 notfound:
	retval = Ns_ReturnNotFound(conn);
    }

 done:
    Ns_DStringFree(&ds);

    return retval;
}


/*
 *----------------------------------------------------------------------
 * FreeEntry --
 *
 *      Logically remove a cached file from file cache.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static void
FreeEntry(void *arg)
{
    File *filePtr = (File *) arg;

    DecrEntry(filePtr);
}


/*
 *----------------------------------------------------------------------
 * DecrEntry --
 *
 *      Decrement reference count of cached file.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static void
DecrEntry(File *filePtr)
{
    if (--filePtr->refcnt == 0) {
	ns_free(filePtr);
    }
}


/*
 *----------------------------------------------------------------------
 * TclDirIndexer --
 *
 *      Default directory indexer.  Runs a Tcl command specified
 *      in the config file.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static int
TclDirIndexer(void *context, Ns_Conn *conn)
{
    int         retval = NS_ERROR;
    Tcl_Interp *interp = Ns_GetConnInterp(conn);

    if (interp != NULL) {
	if (Tcl_Eval(interp, config.tclCommand) == TCL_OK) {
	    retval = NS_OK;
	} else {
	    char message[200];
	    sprintf(message, "\n    (tcl command \"%.150s\" line %d)",
		config.tclCommand, interp->errorLine);
	    Tcl_AddErrorInfo(interp, message);
	    Ns_TclLogError(interp);
	    retval = Ns_ConnReturnInternalError(conn);
	}
    }

    return retval;
}

